/* mfm - A graphical Frontend to mtools                                        */
/* Copyright (c) 1999 Christian Ospelkaus <christian@core-coutainville.org>    */

/* This program is free software; you can redistribute it and/or               */
/* modify it under the terms of the GNU General Public License                 */
/* as published by the Free Software Foundation; either version 2              */
/* of the License, or (at your option) any later version.                      */

/* This program is distributed in the hope that it will be useful,             */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of              */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               */
/* GNU General Public License for more details.                                */

/* You should have received a copy of the GNU General Public License           */
/* along with this program; if not, write to the Free Software                 */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifndef HAVE_CONFIG_H
#include "manualconf.h"
#endif

#include <gtk/gtk.h>
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dirent.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>
#include "mfm.h"

char *removespace(char *string)
{
  int i;
  if(debug)
    printf("In removespace()\n");
  while(string[0]==' ')
    {
      for(i=0;i<strlen(string);i++)
	string[i]=string[i+1];
    }
  for(i=strlen(string)-1;i>=0;i--)
    if(string[i]==' ')
      string[i]='\0';
  return(string);
}

int casecompare(const char *string1,const char *string2)
{
  if(debug)
    printf("In casecompare()\n");
  for(;(*string1)&&(*string2)&&(tolower(*string1)==tolower(*string2));string1++,string2++);
  if((*string1)||(*string2))
    return(0);
  else
    return(1);
}

gint dialog_event(GtkWidget *widget,msgenv *msg)
{
  if(debug)
    {
      printf("In dialog_event()\n");
      printf("  msg: %i\n",msg->msg);
    }
  *(msg->mvar)=msg->msg;
  return(TRUE);
}

gint dialogclose_event(GtkWidget *widget,GdkEvent *event,msgenv *msg)
{
  if(debug)
    {
      printf("In dialogclose_event()\n");
      printf("  msg: %i\n",msg->msg);
    }
  *(msg->mvar)=msg->msg;
  return(TRUE);
}
  
void warnme(char *title,char *text)
{
  GtkWidget *dialog,*okbutton,*label,*hbox;
#ifndef NOPIXMAPS
  GtkWidget *mfmpixwid;
#endif
  int flag=0;
  msgenv msg;
  if(debug)
    printf("In warnme()\n");
  msg.mvar=&flag;
  msg.msg=1;
  dialog=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dialog),title);
  gtk_signal_connect(GTK_OBJECT(dialog),"delete_event",GTK_SIGNAL_FUNC(dialogclose_event),&msg);
  hbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),hbox,TRUE,FALSE,15);
  gtk_widget_show(hbox);

#ifndef NOPIXMAPS
  mfmpixwid=gtk_pixmap_new(mfmpix,mfmmask);
  gtk_box_pack_start(GTK_BOX(hbox),mfmpixwid,FALSE,FALSE,15);
  gtk_widget_show(mfmpixwid);
#endif

  label=gtk_label_new(text);
  gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,15);
  gtk_widget_show (label);

  okbutton=gtk_button_new_with_label("Ok");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),okbutton,TRUE,FALSE,0);
  GTK_WIDGET_SET_FLAGS(okbutton,GTK_CAN_DEFAULT); 
  gtk_widget_grab_default(okbutton);
  gtk_signal_connect(GTK_OBJECT(okbutton),"clicked",GTK_SIGNAL_FUNC(dialog_event),&msg);
  gtk_widget_show(okbutton);

  gtk_grab_add(dialog);
#ifndef OLDGTK
  gtk_window_set_position(GTK_WINDOW(dialog),GTK_WIN_POS_CENTER);
#endif
  gtk_widget_show(dialog);

#ifndef NOPIXMAPS
  gdk_window_set_icon(dialog->window,NULL,mfmpix,mfmmask);
#endif

  while(!flag)
    {
      while (gtk_events_pending())
	gtk_main_iteration();
    }
  gtk_grab_remove(dialog);
  gtk_widget_destroy(dialog);
  if(debug)
    printf("End warnme()\n");
}

int entertext(char *title,char *text,char *answer,int length)
{
  GtkWidget *dialog,*okbutton,*cancelbutton,*label,*hbox,*entry;
#ifndef NOPIXMAPS
  GtkWidget *mfmpixwid;
#endif
  int flag=0;
  msgenv msg[2];

  if(debug)
    printf("In entertext()\n");
  dialog=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dialog),title);
  gtk_signal_connect(GTK_OBJECT(dialog),"delete_event",GTK_SIGNAL_FUNC(dialogclose_event),&msg);
  hbox=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),hbox,TRUE,FALSE,15);
  gtk_widget_show(hbox);

#ifndef NOPIXMAPS
  mfmpixwid=gtk_pixmap_new(mfmpix,mfmmask);
  gtk_box_pack_start(GTK_BOX(hbox),mfmpixwid,FALSE,FALSE,15);
  gtk_widget_show(mfmpixwid);
#endif

  label=gtk_label_new(text);
  gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(hbox),label,FALSE,FALSE,15);
  gtk_widget_show (label);

  entry=gtk_entry_new();
  gtk_entry_set_max_length(GTK_ENTRY(entry),length);
  gtk_entry_set_text(GTK_ENTRY(entry),answer);
  gtk_box_pack_start(GTK_BOX(hbox),entry,FALSE,FALSE,5);
  gtk_widget_show (entry);

  okbutton=gtk_button_new_with_label("Ok");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),okbutton,TRUE,FALSE,0);
  GTK_WIDGET_SET_FLAGS(okbutton,GTK_CAN_DEFAULT); 
  gtk_widget_grab_default(okbutton);
  msg[0].mvar=&flag;
  msg[0].msg=1;
  gtk_signal_connect(GTK_OBJECT(okbutton),"clicked",GTK_SIGNAL_FUNC(dialog_event),msg);
  gtk_widget_show(okbutton);

  cancelbutton=gtk_button_new_with_label("Cancel");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),cancelbutton,TRUE,FALSE,0);
  msg[1].mvar=&flag;
  msg[1].msg=2;
  gtk_signal_connect(GTK_OBJECT(cancelbutton),"clicked",GTK_SIGNAL_FUNC(dialog_event),msg+1);
  gtk_widget_show(cancelbutton);

  gtk_grab_add(dialog);
#ifndef OLDGTK
  gtk_window_set_position(GTK_WINDOW(dialog),GTK_WIN_POS_CENTER);
#endif
  gtk_widget_show(dialog);

#ifndef NOPIXMAPS
  gdk_window_set_icon(dialog->window,NULL,mfmpix,mfmmask);
#endif

  while(!flag)
    {
      while (gtk_events_pending())
	gtk_main_iteration();
    }
  strcpy(answer,gtk_entry_get_text(GTK_ENTRY(entry)));
  gtk_grab_remove(dialog);
  gtk_widget_destroy(dialog);
  if(debug)
    printf("End entertext()\n");
  return(flag);
}

GtkWidget* ambusy(char *title,char *text)
{
  GtkWidget *window,*label,*box;
#ifndef NOPIXMAPS
  GtkWidget *mfmpixwid;
#endif
  int flag=0;
  static msgenv msg;
  msg.mvar=&flag;
  msg.msg=1;
  if(debug)
    printf("In ambusy()\n");

  window = gtk_window_new (GTK_WINDOW_DIALOG);
  gtk_window_set_title (GTK_WINDOW (window), title);
/*   gtk_widget_set_usize(GTK_WIDGET(window), 750, 400); */
  gtk_signal_connect(GTK_OBJECT(window),"delete_event",GTK_SIGNAL_FUNC(dialogclose_event),&msg);
#ifndef OLDGTK
  gtk_container_set_border_width(GTK_CONTAINER(window),15);
#endif

  box=gtk_hbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(window),box);
  gtk_widget_show(box);

#ifndef NOPIXMAPS
  mfmpixwid=gtk_pixmap_new(mfmpix,mfmmask);
  gtk_box_pack_start(GTK_BOX(box),mfmpixwid,FALSE,FALSE,15);
  gtk_widget_show(mfmpixwid);
#endif

  label=gtk_label_new(text);
  gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(box),label,FALSE,FALSE,15);
  gtk_widget_show (label);

  gtk_grab_add(window);
#ifndef OLDGTK
  gtk_window_set_position(GTK_WINDOW(window),GTK_WIN_POS_CENTER);
#endif
  gtk_widget_show(window);

#ifndef NOPIXMAPS
  gdk_window_set_icon(window->window,NULL,mfmpix,mfmmask);
#endif
  if(debug)
    printf("End ambusy()\n");
  return(window);
}

int pathlength(side *whichside)
{
  int i;
  int n=1;
  if(debug)
    printf("In pathlength()\n");
  for(i=0;i<whichside->length;i++)
    n+=strlen(whichside->path[i])+1;
  if(debug)
    {
      printf("  n: %i\n",n);
      printf("End pathlength()\n");
    }
  return(n);
}

void getpath(side *whichside, char **path)
{
  int i;
  if(debug)
    printf("In getpath()\n");
  *path=(char *)malloc((pathlength(whichside)+258)*sizeof(char));
  strcpy(*path,"/");
  for(i=0;i<whichside->length;i++)
    {
      strcat(*path,whichside->path[i]);
      strcat(*path,"/");
    }
  if(debug)
    {
      printf("  path: %s+\n",*path);
      printf("End getpath()\n");
    }
}

void clearpath(side *whichside)
{
  int i;
  if(debug)
    {
      printf("In clearpath()\n");
      printf("  length: %i\n",whichside->length);
    }
  for(i=0;i<whichside->length;i++)
    free(whichside->path[i]);
  whichside->length=0;
  if(debug)
    printf("End clearpath()\n");
}

void gethdpath(side *whichside)
{
  char *path=NULL;
  char *position;
  char *nextposition;
  if(debug)
    {
      printf("In gethdpath()\n");
      printf("  whichside: %p\n",whichside);
    }
  clearpath(whichside);
  errno=0;
  path=malloc(50*sizeof(char));
  while((!getcwd(path,49))&&(errno==ERANGE))
    path=g_realloc(path,(strlen(path)+50)*sizeof(char));
  strcat(path,"/");
  if(debug)
    printf("  path: %s\n",path);
  position=path+1;
  nextposition=strstr(position,"/");
  while(nextposition&&(whichside->length<DIR_DEPTH-1))
    {
      whichside->path[whichside->length]=malloc(sizeof(char)*((nextposition-position)+1));
      strncpy(whichside->path[whichside->length],position,nextposition-position);
      whichside->path[whichside->length][nextposition-position]='\0';
      whichside->length+=1;
      position=nextposition+1;
      nextposition=strstr(position,"/");
    }
  free(path);
  if(debug)
    {
      printf("  length: %i\n",whichside->length);
      printf("End gethdpath()\n");
    }
  return;
}

int read_configfile(side *leftside, const char *cfname)
{
  int i;
  char buff[120];
  char *pos;
  FILE *cfile;
  if(debug)
    {
      printf("In read_configfile()\n");
      printf("  leftside: %p\n",leftside);
    }
  if((cfile=fopen(cfname,"r")))
    {
      strcpy(drives,"");
      while(fgets(buff,119,cfile))
	{
	  if((pos=strchr(buff,'#')))
	    *pos='\0';
	  if((pos=strchr(buff,'\n')))
	    *pos='\0';
	  if(!strncmp(buff,"DRIVES=\"",8))
	    for(i=8;(pos=strchr("abcdefghijklmnopqrstuvwxyz",buff[i]));i++)
	      if(!strchr(drives,*pos))
		{
		  drives[strlen(drives)+1]='\0';
		  drives[strlen(drives)]=*pos;
		}
	  if(!strncmp(buff,"MPATH=\"",7))
	    {
	      for(i=0;(buff[i+7]!='\"')&&buff[i+7]&&(buff[i+7]!='\n');i++)
		mpath[i]=buff[i+7];
	      mpath[i]='\0';
	    }
	  if(!strncmp(buff,"LEFTDRIVE=\"",11))
	    leftside->drive=buff[11];
	  if(!strncmp(buff,"RIGHTDRIVE=\"",12))
	    leftside->other->drive=buff[12];
	}
      fclose(cfile);
      if(!strchr(drives,leftside->drive))
	leftside->drive='\0';
      if(!strchr(drives,leftside->other->drive))
	leftside->other->drive='\0';
      if(debug)
	{
	  printf("  found %s\n",cfname);
	  printf("  drives %s\n",drives);
	  printf("  leftside->drive %c\n",leftside->drive);
	  printf("  leftside->other->drive %c\n",leftside->other->drive);
	  printf("End read_configfile()\n");
	}
      return(0);
    }
  else
    {
      if(debug)
	{
	  printf("  didnt find %s\n",cfname);
	  printf("End read_configfile()\n");
	}
      return(1);
    }
}

void save_config(side *leftside)
{
  char *cfname;
  FILE *cfile;
  cfname=malloc((strlen(homedir)+6)*sizeof(char));
  strcpy(cfname,homedir);
  strcat(cfname,"/.mfm");
  if(debug)
    printf("In save_config()\n");
  if((cfile=fopen(cfname,"w")))
    {
      fprintf(cfile,"# mfm config file. comments start with a hash sign.\n");
      fprintf(cfile,"# \n");
      fprintf(cfile,"# This variable sets the allowed driveletters (all lowercase). Example:\n");
      fprintf(cfile,"# DRIVES=\"ab\"\n");
      fprintf(cfile,"DRIVES=\"%s\"\n",drives);
      fprintf(cfile,"# \n");
      fprintf(cfile,"# This variable sets the driveletter upon startup in the left window. \n");
      fprintf(cfile,"# An empty string or space is for the hardisk. Example: \n");
      fprintf(cfile,"# LEFTDRIVE=\"a\"\n");
      fprintf(cfile,"LEFTDRIVE=\"%c\"\n",leftside->drive?leftside->drive:' ');
      fprintf(cfile,"# \n");
      fprintf(cfile,"# This variable sets the driveletter upon startup in the right window. \n");
      fprintf(cfile,"# An empty string or space is for the hardisk. Example: \n");
      fprintf(cfile,"# RIGHTDRIVE=\"a\"\n");
      fprintf(cfile,"RIGHTDRIVE=\"%c\"\n",leftside->other->drive?leftside->other->drive:' ');
      fprintf(cfile,"# This variable sets the path to the mtools commands. \n");
      fprintf(cfile,"# If left empty, " MTOOLSDIR " is assumed.\n");
      fprintf(cfile,"MPATH=\"%s\"\n",mpath);
      fclose(cfile);
    }
  free(cfname);
}
  
void set_combo_entries(side *whichside)
{
  char *pos,*buff;
  char curdrive[3]=" :";
  GList *listpos,*bufflist;
  if(debug)
    printf("In set_combo_entries()\n");
  listpos=NULL;
  listpos=g_list_append(listpos,"Harddisk");
  for(pos=drives;*pos;pos++)
    {
      buff=malloc(3*sizeof(char));
      buff[0]=*pos;
      buff[1]=':';
      buff[2]='\0';
      listpos=g_list_append(listpos,buff);
    }
  bufflist=whichside->comboglist;
  gtk_combo_set_popdown_strings(GTK_COMBO(whichside->combo),listpos);
  for(listpos=bufflist;listpos;listpos=listpos->next)
    free(listpos->data);
  g_list_free(bufflist);
  if(!whichside->drive)
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(whichside->combo)->entry),"Harddisk");
  else
    {
      curdrive[0]=whichside->drive;
      gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(whichside->combo)->entry),curdrive);
    }
  if(debug)
    printf("End set_combo_entries()\n");
}

int delete_recursive(const char *path)
{
  DIR *directory;
  char *fullpath;
  struct dirent *entry;
  struct stat buf;
  int result=0;

  if(debug)
    printf("In delete_recursive()\n");
  fullpath=malloc((strlen(path)+258)*sizeof(char));
  directory=opendir(path);
  if(directory)
    {
      entry=readdir(directory);
      while(entry)
	{
	  strcpy(fullpath,path);
	  strcat(fullpath,"/");
	  strcat(fullpath,entry->d_name);
	  if(strcmp(entry->d_name,"..")&&strcmp(entry->d_name,"."))
	    {
	      stat(fullpath,&buf);
	      if(S_ISDIR(buf.st_mode))
		{
		  if(delete_recursive(fullpath))
		    {
		      break;
		      result=1;
		    }
		}
	      else
		{
		  if(unlink(fullpath))
		    {
		      break;
		      result=1;
		    }
		} 
	    }
	  entry=readdir(directory);
	}
      if(rmdir(path))
	result=1;
    }
  free(fullpath);
  if(debug)
    printf("End delete_recursive()\n");
  return(result);
}
    
void destroy_dentry(gpointer data)
{
  if(debug)
    printf("In destroy_dentry()\n");
  free(data);
  if(debug)
    printf("End destroy_dentry()\n");
}

void selectall_event(GtkWidget *widget,side *whichside) 
{
  if(debug)
    printf("In selectall_event()\n");
  if (!GTK_CLIST(whichside->list)->rows) return;
  gtk_clist_select_all(GTK_CLIST(whichside->list));
}

void selectnone_event(GtkWidget *widget,side *whichside) 
{
  if(debug)
    printf("In selectnone_event()\n");
  gtk_clist_unselect_all(GTK_CLIST(whichside->list));
}

void refresh_event(GtkWidget *widget,side *whichside) 
{
  char *path;
  dentry *current;

  if(debug)
    printf("In refresh_event()\n");

  gtk_clist_clear(GTK_CLIST(whichside->list));

  getpath(whichside,&path);
  if(!whichside->drive)
    {
      DIR *directory;
      struct stat buf;
      struct dirent *entry;
      char *position;
      char size[16]="";
      char date[11]="";
      char time[6]="";
      int pathlen;
      gint row;
      gchar *tmppoint[5];
      if(debug)
	printf("  reading from unix-filesystem...\n");
      tmppoint[0]=NULL;
      tmppoint[2]=(gchar *) &size;
      tmppoint[3]=(gchar *) &date;
      tmppoint[4]=(gchar *) &time;
      pathlen=pathlength(whichside);
      position=path+strlen(path);
      gtk_statusbar_pop(GTK_STATUSBAR(whichside->status),whichside->context);
      gtk_statusbar_push(GTK_STATUSBAR(whichside->status),whichside->context,path);
      directory=opendir(path);
      if(directory)
	{
	  gtk_clist_freeze(GTK_CLIST(whichside->list));
	  if(debug)
	    printf("  entering loop and reading directory contents...\n");
	  entry=readdir(directory);
	  while(entry)
	    {
	      strcpy(position,entry->d_name);
	      if(debug)
		printf("  filename: %s+\n",path);
	      if( (GTK_CHECK_MENU_ITEM(showdots)->active||!strcmp(position,"..")||(*position!='.'))
		  &&strcmp(position,".")
		  &&(!((pathlen==1)&&(!strcmp(position,"..")))))
		{
		  current=malloc(sizeof(dentry));
		  strcpy(current->name,entry->d_name);
		  stat(path,&buf);
		  if(S_ISDIR(buf.st_mode))
		    current->dir=1;
		  else
		    current->dir=0;
		  tmppoint[1]=current->name;
		  sprintf(size,"%li",buf.st_size);
		  current->size=buf.st_size;
		  sprintf(date,"%-2.2i-%-2.2i-%-4.4i",
			  localtime(&(buf.st_ctime))->tm_mon,
			  localtime(&(buf.st_ctime))->tm_mday,
			  localtime(&(buf.st_ctime))->tm_year+1900);
		  sprintf(time,"%-2.2i:%-2.2i",
			  localtime(&(buf.st_ctime))->tm_hour,
			  localtime(&(buf.st_ctime))->tm_min);
		  row=gtk_clist_append(GTK_CLIST(whichside->list),tmppoint);
		  if(!strcmp(position,".."))
		    gtk_clist_set_selectable(GTK_CLIST(whichside->list),row,FALSE);
		  lstat(path,&buf);
		  if(current->dir)
		    {
		      if(S_ISLNK(buf.st_mode))
			{
			  gtk_clist_set_foreground(GTK_CLIST(whichside->list),row,whichside->linkcolor);
#ifndef NOPIXMAPS
			  gtk_clist_set_pixmap(GTK_CLIST(whichside->list),row,0,linkdirpix,linkdirmask);
#endif
			}
		      else
			{
			  gtk_clist_set_foreground(GTK_CLIST(whichside->list),row,whichside->dircolor);
#ifndef NOPIXMAPS
			  gtk_clist_set_pixmap(GTK_CLIST(whichside->list),row,0,dirpix,dirmask);
#endif
			}
		    }
		  else
		    {
		      if(S_ISLNK(buf.st_mode))
			{
			  gtk_clist_set_foreground(GTK_CLIST(whichside->list),row,whichside->linkcolor);
#ifndef NOPIXMAPS
			  gtk_clist_set_pixmap(GTK_CLIST(whichside->list),row,0,linkfilepix,linkfilemask);
#endif
			}
		      else
			{
#ifndef NOPIXMAPS
			  gtk_clist_set_pixmap(GTK_CLIST(whichside->list),row,0,filepix,filemask);
#endif
			}
		    }
		  gtk_clist_set_row_data_full(GTK_CLIST(whichside->list),row,(gpointer)current,destroy_dentry);
		}
	      entry=readdir(directory);
	    }
	  closedir(directory);
	  if(debug)
	    printf("  end of loop\n");
	  gtk_clist_thaw(GTK_CLIST(whichside->list));
	}
      else if(whichside->length)
	{
	  if(debug)
	    {
	      printf("  Opening of directory not successfull.\n");
	      printf("  cd to parent and recurse.\n");
	    }
	  whichside->length-=1;
	  free(whichside->path[whichside->length]);
	  free(path);
	  refresh_event(whichside->list,whichside);
	  if(debug)
	    printf("End refresh_event()\n");
	  return;
	}
    }
  else
    {
      FILE *outpipe;
      char buff[340];
      char *command;
      char dosname[9];
      char dosext[4];
      char size[10];
      char date[11];
      char time[6];
      char vfatname[298];
      int entcount=0;
      void *success;
      gint row;
      gchar *tmppoint[5];
/*       GtkWidget *busywindow; */
      if(debug)
	printf("  reading from floppy...\n");
/*       busywindow=ambusy("working...","Reading directory contents. Please wait."); */
      tmppoint[0]=NULL;
      tmppoint[1]=(gchar *) &vfatname;
      tmppoint[2]=(gchar *) &size;
      tmppoint[3]=(gchar *) &date;
      tmppoint[4]=(gchar *) &time;
      command=(char *) malloc((strlen(path)+strlen(mpath)+11)*sizeof(char));
      strcpy(command,mpath);
      strcat(command,"/mdir \" :");
      command[strlen(mpath)+7]=whichside->drive;
      strcat(command,path);
      gtk_statusbar_pop(GTK_STATUSBAR(whichside->status),whichside->context);
      gtk_statusbar_push(GTK_STATUSBAR(whichside->status),whichside->context,command+strlen(mpath)+7);
      strcat(command,"\"");
      if(debug)
	printf("  command: %s+\n",command);
      if ((outpipe=popen(command,"r"))==NULL)
	{
	  g_print("Failure.\n");
	  gtk_main_quit();
	}
      free(command);
      while((success=fgets(buff,340,outpipe))
	    &&strncmp(buff," Directory for",14)
	    &&strncmp(buff,"Directory for",13));
      if(!success)
	{
	  if(debug)
	    printf("  could not get \"Directory for\"\n");
	  pclose(outpipe);
/* 	  gtk_grab_remove(busywindow); */
/* 	  gtk_widget_destroy(busywindow); */
	  if(whichside->length)
	    {
	      if(debug)
		printf("  trying again...\n");
	      clearpath(whichside);
	      refresh_event(whichside->list,whichside);
	    }
	  free(path);
	  return;
	}
      if(!fgets(buff,340,outpipe)||(strlen(buff)!=1))
	{
	  if(debug)
	    printf("  unexpected output\n");
	  pclose(outpipe);
/* 	  gtk_grab_remove(busywindow); */
/* 	  gtk_widget_destroy(busywindow); */
	  if(whichside->length)
	    {
	      if(debug)
		printf("  trying again...\n");
	      clearpath(whichside);
	      refresh_event(whichside->list,whichside);
	    }
	  free(path);
	  return;
	}
      if(debug)
	printf("  entering loop and reading directory contents...\n");
      gtk_clist_freeze(GTK_CLIST(whichside->list));
      while(fgets(buff,340,outpipe)&&(buff[0]!=' ')&&(strlen(buff)>=40))
	{
	  if(debug)
	    printf("  read line: %s+\n",buff);
	  entcount+=1;
	  strncpy(dosname,buff,8);
	  strncpy(dosext,buff+9,3);
	  strncpy(size,buff+13,9);
	  strncpy(date,buff+23,10);
	  strncpy(time,buff+35,5);
	  if(*time==' ')
	    *time='0';
	  dosname[8]='\0';
	  dosext[3]='\0';
	  size[9]='\0';
	  date[10]='\0';
	  time[5]='\0';
	  removespace(dosname);
	  removespace(dosext);
	  removespace(size);
	  removespace(time);
	  if(strlen(buff)>42)
	    {
	      strncpy(vfatname,buff+42,strlen(buff)-42);
	      vfatname[strlen(buff+42)-1]='\0';
	    }
	  else
	    {
	      strcpy(vfatname,dosname);
	      if(strlen(dosext))
		{
		  strcat(vfatname,".");
		  strcat(vfatname,dosext);
		}
	    }
	  if(strcmp(vfatname,"."))
	    {
	      current=malloc(sizeof(dentry));
	      strcpy(current->name,vfatname);
	      if(strcmp(size,"<DIR>"))
		current->dir=0;
	      else
		{
		  current->dir=1;
		  strcpy(size,"0");
		}
	      current->size=atol(size);
	      row=gtk_clist_append(GTK_CLIST(whichside->list),tmppoint);
	      if(!strcmp(vfatname,".."))
		gtk_clist_set_selectable(GTK_CLIST(whichside->list),row,FALSE);
#ifndef NOPIXMAPS
	      if(current->dir)
		gtk_clist_set_pixmap(GTK_CLIST(whichside->list),row,0,dirpix,dirmask);
	      else
		gtk_clist_set_pixmap(GTK_CLIST(whichside->list),row,0,filepix,filemask);
#endif
	      gtk_clist_set_row_data_full(GTK_CLIST(whichside->list),row,(gpointer)current,destroy_dentry);
	    }
	}
      if(debug)
	printf("  end of loop\n");
      if(!entcount&&whichside->length)
	{
	  if(debug)
	    printf("  strange output: zero entries in subdir\n");
	  clearpath(whichside);
	  gtk_clist_thaw(GTK_CLIST(whichside->list));
	  if(debug)
	    printf("  trying again...\n");
	  refresh_event(whichside->list,whichside);
	}
      else
	gtk_clist_thaw(GTK_CLIST(whichside->list));
      pclose(outpipe);
/*       gtk_grab_remove(busywindow); */
/*       gtk_widget_destroy(busywindow); */
    }
  free(path);
  gtk_clist_sort(GTK_CLIST(whichside->list));
  if(debug)
    printf("End refresh_event()\n");
  return;
}

void delete_event(GtkWidget *widget,GdkEvent *event,side *whichside)
{
  if(debug)
    printf("In delete_event()\n");
  if(!nosave)
    save_config(whichside);
  if(debug)
    printf("End delete_event()\n");
  gtk_main_quit ();
}

void quit_event(GtkWidget *widget,side *whichside)
{  
  if(debug)
    printf("In quit_event()\n");
  if(!nosave)
    save_config(whichside);
  if(debug)
    printf("End quit_event()\n");
  gtk_main_quit ();
}

void copy_event(GtkWidget *widget,side *from)
{
  side *to=from->other;
  GList *items;
  gint row;
  msgenv msg;
  GtkWidget *progressdialog,*progressbar,*progresslabel,*cancelbutton;
  long copiedbytes=0;
  char *text,*buff,*frompath,*topath,*name1,*name2,*cpath;
  char warning[400];
  int status,i,j,flag=0;
  pid_t childpid;

  if(debug)
    printf("In copy_event()\n");

  msg.mvar=&flag;
  msg.msg=1;
  getpath(from,&frompath);
  getpath(to,&topath);

  cpath=(char*)malloc((strlen(mpath)+7)*sizeof(char));
  strcpy(cpath,mpath);
  strcat(cpath,"/mcopy");

  name1=malloc((strlen(frompath)+strlen(topath)+258)*sizeof(char));
  name2=malloc((strlen(frompath)+strlen(topath)+258)*sizeof(char));

  if(debug)
    printf("  creating dialog\n");
  progressdialog=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(progressdialog),"Copying files. Please wait.");
  gtk_widget_set_usize(GTK_WIDGET(progressdialog),300,0);
  gtk_signal_connect(GTK_OBJECT(progressdialog),"delete_event",GTK_SIGNAL_FUNC(dialogclose_event),&msg);
#ifndef OLDGTK
  gtk_window_set_position(GTK_WINDOW(progressdialog),GTK_WIN_POS_CENTER);
  gtk_container_set_border_width(GTK_CONTAINER(progressdialog),10);
#endif
  
  progressbar=gtk_progress_bar_new();
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(progressdialog)->vbox),progressbar,TRUE,TRUE,5);
  gtk_widget_show(progressbar);

  progresslabel=gtk_label_new("Copying files...\n\nThis progressbar will be updated\nafter each copied file.");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(progressdialog)->vbox),progresslabel,TRUE,TRUE,5);
  gtk_widget_show(progresslabel);
  
  cancelbutton=gtk_button_new_with_label("Cancel");
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(progressdialog)->action_area),cancelbutton,TRUE,FALSE,0);
  gtk_signal_connect(GTK_OBJECT(cancelbutton),"clicked",GTK_SIGNAL_FUNC(dialog_event),&msg);
  gtk_widget_show(cancelbutton);
  GTK_WIDGET_SET_FLAGS(cancelbutton,GTK_CAN_DEFAULT); 
  gtk_widget_grab_default(cancelbutton);

  gtk_widget_show(progressdialog);
#ifndef NOPIXMAPS
  gdk_window_set_icon(progressdialog->window,NULL,mfmpix,mfmmask);
#endif
  gtk_grab_add(progressdialog);

  if(debug)
    printf("  entering loop...\n");
  items=GTK_CLIST(from->list)->selection;
  while (items&&(!flag))
    {
      row=(gint)items->data;
      gtk_clist_get_text(GTK_CLIST(from->list),row,1,&text);
      i=0;
      j=gtk_clist_get_text(GTK_CLIST(to->list),i,1,&buff);
      while(j)
	{
	  if(casecompare(buff,text))
	    {
	      i=-1;
	      break;
	    }
	  i+=1;
	  j=gtk_clist_get_text(GTK_CLIST(to->list),i,1,&buff);
	}
      if(i==-1)
	{
	  if(debug)
	    printf("  file exists\n");
	  strcpy(warning,"The file\n\n");
	  strcat(warning,text);
	  strcat(warning,"\n\nseems to exist in the target directory. If you want to overwrite it,\n\
delete it in the target directory first and copy it again.");
	  warnme("File exists",warning);
	  items=items->next;
	  continue;
	}
      if(from->drive)
	{
	  name1[0]=from->drive;
	  name1[1]=':';
	  name1[2]='\0';
	}
      else
	name1[0]='\0';
      strcat(name1,frompath);
      strcat(name1,text);
      if(to->drive)
	{
	  name2[0]=to->drive;
	  name2[1]=':';
	  name2[2]='\0';
	}
      else
	name2[0]='\0';
      strcat(name2,topath);
      if((childpid=fork())<0)
	{
	  printf("  Error with fork()\n");
	  exit(-1);
	}
      else if(childpid==0) /* child */
	{
	  if(debug)
	    {
	      printf("  in child\n");
	      printf("  cpath: %s+\n",cpath);
	      printf("  name1: %s+\n",name1);
	      printf("  name2: %s+\n",name2);
	    }
	  if(GTK_CHECK_MENU_ITEM(textmode)->active)
	    execl(cpath,"mcopy","-nmot",name1,name2,NULL);
	  else
	    execl(cpath,"mcopy","-nmo",name1,name2,NULL);
	  printf("  error in execl()\n");
	  printf("  errno:%d\n",errno);
	  _exit(-1); 
	}
      else /* parent */
	{
	  if(debug)
	    printf("  in parent\n");
	  while(1) 
	    {
	      while (gtk_events_pending())
		gtk_main_iteration();
	      if(waitpid(childpid, &status, WNOHANG) == childpid)
		{
		  waitpid(childpid, &status, WUNTRACED);
		  break;
		}
	    }
	}
      copiedbytes+=((dentry *)gtk_clist_get_row_data(GTK_CLIST(from->list),(int)(items->data)))->size;
      if(from->totsize)
	gtk_progress_bar_update(GTK_PROGRESS_BAR(progressbar),(float) copiedbytes/(from->totsize));
      items=items->next;
    }
  gtk_grab_remove(progressdialog);
  gtk_widget_destroy(progressdialog);

  if(debug)
    printf("  freeing memory...\n");
  free(frompath);
  free(topath);
  free(name1);
  free(name2);
  if(debug)
    printf("  calling refresh_event\n");
  refresh_event(to->list,to);
  if(debug)
    printf("End copy_event()\n");
}

void config_event(GtkWidget *widget,side *leftside)
{
  GtkWidget *dialog,*button,*entrydrive,*entrympath,*box;
  GtkWidget *label;
  msgenv msg[2];
  int flag=0;
  int i;
  char *buttonlabels[]={"OK","Cancel"};
  char newdrives[27];
  char *pos;
  if(debug)
    printf("In config_event()\n");
  dialog=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dialog),"Configure mfm");
#ifndef OLDGTK
  gtk_container_set_border_width(GTK_CONTAINER(dialog),10);
#endif

  /* Display help */
  box=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),box,TRUE,TRUE,5);
  gtk_widget_show(box);

  label=gtk_label_new(CONFDRIVES);
  gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(box),label,FALSE,FALSE,0);
  gtk_widget_show(label);

  /* Create textfield for driveletters */
  box=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),box,TRUE,TRUE,5);
  gtk_widget_show(box);

  entrydrive=gtk_entry_new_with_max_length(26);
  gtk_widget_set_usize(entrydrive,150,0);
  gtk_entry_set_text(GTK_ENTRY(entrydrive),drives);
  gtk_entry_select_region(GTK_ENTRY(entrydrive),0,strlen(drives));
  gtk_box_pack_start(GTK_BOX(box),entrydrive,TRUE,FALSE,0);
  gtk_widget_show (entrydrive);

  /* Display help */
  box=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),box,TRUE,TRUE,5);
  gtk_widget_show(box);

  label=gtk_label_new(CONFPATH);
  gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(box),label,TRUE,TRUE,0);
  gtk_widget_show (label);

  /* Create textfield for path */
  box=gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),box,TRUE,TRUE,5);
  gtk_widget_show(box);

  entrympath=gtk_entry_new_with_max_length(255);
  gtk_widget_set_usize(entrympath,150,0);
  gtk_entry_set_text(GTK_ENTRY(entrympath),mpath); 
/*   gtk_entry_select_region(GTK_ENTRY(entrympath),0,strlen(mpath)); */
  gtk_box_pack_start(GTK_BOX(box),entrympath,TRUE,FALSE,0);
  gtk_widget_show (entrympath);

  /* Create buttons */
  for(i=0;i<=1;i++)
    {
      msg[i].mvar=&flag;
      msg[i].msg=i+1;
      button=gtk_button_new_with_label(buttonlabels[i]);
      gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),button,TRUE,TRUE,0);
      gtk_signal_connect(GTK_OBJECT(button),"clicked",GTK_SIGNAL_FUNC(dialog_event),msg+i);
      if(!i)
	{
	  GTK_WIDGET_SET_FLAGS(button,GTK_CAN_DEFAULT); 
	  gtk_widget_grab_default(button);
	}
      gtk_widget_show(button);
    }
  gtk_signal_connect(GTK_OBJECT(dialog),"delete_event",GTK_SIGNAL_FUNC(dialogclose_event),msg+1);

  gtk_grab_add(dialog);
#ifndef OLDGTK
  gtk_window_set_position(GTK_WINDOW(dialog),GTK_WIN_POS_CENTER);
#endif
  gtk_widget_show(dialog);

#ifndef NOPIXMAPS
  gdk_window_set_icon(dialog->window,NULL,mfmpix,mfmmask);
#endif

  while(!flag)
    {
      while (gtk_events_pending())
	gtk_main_iteration();
    }
  if(flag==1)
    {
      strcpy(newdrives,gtk_entry_get_text(GTK_ENTRY(entrydrive)));
      strcpy(drives,"");
      for(pos=newdrives;*pos;pos++)
	if(strchr("abcdefghijklmnopqrstuvwxyz",*pos))
	  {
	    drives[strlen(drives)+1]='\0';
	    drives[strlen(drives)]=*pos;
	  }
      strcpy(mpath,gtk_entry_get_text(GTK_ENTRY(entrympath)));
      if(!strchr(drives,leftside->drive)&&leftside->drive)
	leftside->drive='\0';
      if(!strchr(drives,leftside->other->drive)&&leftside->other->drive)
	leftside->other->drive='\0';
      set_combo_entries(leftside);
      set_combo_entries(leftside->other);
    }
  gtk_grab_remove(dialog);
  gtk_widget_destroy(dialog);
  if(debug)
    printf("End config_event()\n");
}

int ask_delete(const char *string1,const char *string2,const char *string3)
{
  GtkWidget *dialog,*button;
  GtkWidget *label;
  msgenv msg[4];
  int result=0;
  int i;
  char *buttonlabels[]={"Yes","Skip","All","Cancel"};
  if(debug)
    printf("In ask_delete()\n");
  dialog=gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dialog),"Really delete files?");
  gtk_widget_set_usize(GTK_WIDGET(dialog),300,200);

  /* Display warning */
  label=gtk_label_new(string1);
  gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),label,TRUE,TRUE,0);
  gtk_widget_show (label);

  label=gtk_label_new(string2);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),label,TRUE,TRUE,0);
  gtk_widget_show (label);

  label=gtk_label_new(string3);
  gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_LEFT);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),label,TRUE,TRUE,0);
  gtk_widget_show (label);

  /* Create buttons */
  for(i=0;i<=3;i++)
    {
      msg[i].mvar=&result;
      msg[i].msg=i+1;
      button=gtk_button_new_with_label(buttonlabels[i]);
      gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),button,TRUE,TRUE,0);
      gtk_signal_connect(GTK_OBJECT(button),"clicked",GTK_SIGNAL_FUNC(dialog_event),msg+i);
      gtk_widget_show(button);
    }
  gtk_signal_connect(GTK_OBJECT(dialog),"delete_event",GTK_SIGNAL_FUNC(dialogclose_event),msg+3);

  gtk_grab_add(dialog);
#ifndef OLDGTK
  gtk_window_set_position(GTK_WINDOW(dialog),GTK_WIN_POS_CENTER);
#endif
  gtk_widget_show(dialog);

#ifndef NOPIXMAPS
  gdk_window_set_icon(dialog->window,NULL,mfmpix,mfmmask);
#endif

  while(!result)
    {
      while (gtk_events_pending())
	gtk_main_iteration();
    }
  gtk_grab_remove(dialog);
  gtk_widget_destroy(dialog);
  if(debug)
    printf("End ask_delete()\n");
  return(result);
}

void filedelete_event(GtkWidget *widget,side *whichside)
{
  GList *items;
  gint row;
  int allflag=0;
  int allrflag=0;
  int askresult;
  char *text;
  char *path;
  char *command;
  FILE *outpipe;

  if(debug)
    {
      printf("In filedelete_event()\n");
      printf("  whichside %p\n",whichside);
    }

  getpath(whichside,&path);

  command=malloc((strlen(path)+strlen(mpath)+268)*sizeof(char));
  items=GTK_CLIST(whichside->list)->selection;
  while (items)
    {
      row=(gint)items->data;
      gtk_clist_get_text(GTK_CLIST(whichside->list),row,1,&text);
      if(((dentry *)gtk_clist_get_row_data(GTK_CLIST(whichside->list),row))->dir)
	{
	  strcpy(command,mpath);
	  strcat(command,"/mdeltree \" :");
	  command[strlen(mpath)+11]=whichside->drive;
	  strcat(command,path);
	  strcat(command,text);
	  if(whichside->drive)
	    {
	      askresult=2;
	      if(!allrflag)
		askresult=ask_delete("",command+strlen(mpath)+11,"\nis a directory. Do you want to remove this directory\nrecursively with all its subdirectories\?");
	      if((askresult==1)||(askresult==3)||allrflag)
		{
		  strcat(command,"\"");
		  outpipe=popen(command,"r");
		  pclose(outpipe);
		}
	      if(askresult==3)
		allrflag=1;
	      if(askresult==4)
		break;
	    }
	  else
	    {
	      askresult=2;
	      if(!allrflag)
		askresult=ask_delete("",command+strlen(mpath)+11,"\nis a directory. Do you want to remove this directory\nrecursively with all its subdirectories\?");
	      if((askresult==1)||(askresult==3)||allrflag)
		delete_recursive(command+strlen(mpath)+11);
	      if(askresult==3)
		allrflag=1;
	      if(askresult==4)
		break;
	    }
	}
      else
	{
	  strcpy(command,mpath);
	  strcat(command,"/mdel \" :");
	  command[strlen(mpath)+7]=whichside->drive;
	  strcat(command,path);
	  strcat(command,text);
	  if(whichside->drive)
	    {
	      askresult=2;
	      if(!allflag)
		askresult=ask_delete("Deleting file\n\n",command+strlen(mpath)+7,"\nDo you really want to do this\?");
	      if((askresult==1)||(askresult==3)||allflag)
		{
		  strcat(command,"\"");
		  outpipe=popen(command,"r");
		  pclose(outpipe);
		}
	      if(askresult==3)
		allflag=1;
	      if(askresult==4)
		break;
	    }
	  else
	    {
	      askresult=2;
	      if(!allflag)
		askresult=ask_delete("Deleting file\n\n",command+strlen(mpath)+7,"\nDo you really want to do this\?");
	      if((askresult==1)||(askresult==3)||allflag)
		{
		  unlink(command+strlen(mpath)+7);
		}
	      if(askresult==3)
		allflag=1;
	      if(askresult==4)
		break;
	    }
	}
      items=items->next;
    }
  free(path);
  if(debug)
    printf("  calling refresh_event()\n");
  refresh_event(whichside->list,whichside);
  refresh_event(whichside->other->list,whichside->other);
  if(debug)
    printf("End filedelete_event()\n");
}

void filerename_event(GtkWidget *widget,side *whichside)
{
  GList *items;
  gint row;
  int flag;
  char *text;
  char *path;
  char *command;
  char *hello;
  char newname[260];
  FILE *outpipe;

  if(debug)
    printf("In filerename_event()\n");

  getpath(whichside,&path);

  command=malloc((2*strlen(path)+strlen(mpath)+2*268)*sizeof(char));
  hello=malloc((260+strlen(path))*sizeof(char));
  items=GTK_CLIST(whichside->list)->selection;
  while(items)
    {
      row=(gint)items->data;
      gtk_clist_get_text(GTK_CLIST(whichside->list),row,1,&text);
      strcpy(command,mpath);
      strcat(command,"/mren \" :");
      command[strlen(mpath)+7]=whichside->drive;
      strcat(command,path);
      strcat(command,text);
      strcpy(hello,"Enter new name for \"");
      strcat(hello,text);
      strcat(hello,"\" :");
      strcpy(newname,text);
      if((flag=entertext("Rename file / directory",hello,newname,255))==1)
	{
	  if(whichside->drive)
	    {
	      strcat(command,"\" \" ");
	      command[strlen(command)-1]=whichside->drive;
	      strcat(command,":");
	      strcat(command,path);
	      strcat(command,newname);
	      strcat(command,"\"");
	      outpipe=popen(command,"r");
	      pclose(outpipe);
	    }
	  else
	    {
	      strcpy(hello,path);
	      strcat(hello,newname);
	      rename(command+strlen(mpath)+7,hello);
	    }
	}
      items=items->next;
    }
  free(path);
  free(hello);
  refresh_event(whichside->list,whichside);
  if(debug)
    printf("End filerename_event()\n");
}

void fileprint_event(GtkWidget *widget,side *whichside)
{
  GList *items;
  gint row;
  char *text;
  char *path;
  char *command;
  char lprcommand[21]="lpr";

  if(debug)
    printf("In fileprint_event()\n");

  if(entertext("Print files","Enter printer command",lprcommand,20)==1)
    {
      getpath(whichside,&path);
      command=malloc((strlen(path)+strlen(mpath)+300)*sizeof(char));
      items=GTK_CLIST(whichside->list)->selection;
      while(items)
	{
	  row=(gint)items->data;
	  gtk_clist_get_text(GTK_CLIST(whichside->list),row,1,&text);
	  if(!((dentry *)gtk_clist_get_row_data(GTK_CLIST(whichside->list),row))->dir)
	    {
	      if(whichside->drive)
		{
		  strcpy(command,mpath);
		  strcat(command,"/mtype \" :");
		  command[strlen(mpath)+8]=whichside->drive;
		  strcat(command,path);
		  strcat(command,text);
		  strcat(command,"\" | ");
		  strcat(command,lprcommand);
		  system(command);
		}
	      else
		{
		  strcpy(command,lprcommand);
		  strcat(command," \"");
		  strcat(command,path);
		  strcat(command,text);
		  strcat(command,"\"");
		  system(command);
		}
	    }
	  items=items->next;
	}    
      free(path);
      free(command);
    }
  if(debug)
    printf("End fileprint_event()\n");
}

void mkdir_event(GtkWidget *widget,side *whichside)
{
  int flag;
  char *path;
  char *command;
  char newname[260];
  FILE *outpipe;

  if(debug)
    printf("In mkdir_event()\n");

  getpath(whichside,&path);

  strcpy(newname,"New directory");
  command=malloc((2*strlen(path)+strlen(mpath)+2*268)*sizeof(char));
  strcpy(command,mpath);
  strcat(command,"/mmd \" :");
  command[strlen(mpath)+6]=whichside->drive;
  strcat(command,path);
  if((flag=entertext("Make directory","Enter new name for new directory:",newname,255))==1)
    {
      strcat(command,newname);
      if(whichside->drive)
	{
	  strcat(command,"\"");
	  outpipe=popen(command,"r");
	  pclose(outpipe);
	}
      else
	{
	  mkdir(command+strlen(mpath)+6,0777);
	}
    }
  free(path);
  refresh_event(whichside->list,whichside);
  if(debug)
    printf("End mkdir_event()\n");
}

void about_event(GtkWidget *widget,GdkEvent *event,gpointer Data)
{
  if(debug)
    printf("In about_event()\n");
  warnme("about mfm",ABOUTMFM);
}

void use_event(GtkWidget *widget,GdkEvent *event,gpointer Data)
{
  if(debug)
    printf("In use_event()\n");
  warnme("How to use mfm",MFMUSE);
}

void command_event(GtkWidget *widget,GdkEvent *event,gpointer Data)
{
  if(debug)
    printf("In command_event()\n");
  warnme("Command line options",MFMOPTIONS);
}

void combo_change(GtkWidget *widget,side *whichside) 
{
  char *string;
  if(debug)
    printf("In combo_change()\n");
  string=gtk_entry_get_text(GTK_ENTRY(widget));
  if(strcmp(string,"Harddisk"))
    {
      whichside->drive=string[0];
      clearpath(whichside);
    }
  else
    {
      whichside->drive='\0';
      clearpath(whichside);
      gethdpath(whichside);
    }
  refresh_event(whichside->list,whichside);
  if(debug)
    printf("End combo_change()\n");
  return;
}

gint dclick_callback(GtkWidget *widget,gint row,gint column,GdkEventButton *event,side *whichside)
{
  long totsize=0;
  GList *items;
  char *path,*drivepath;

  if(debug)
    printf("In dclick_callback()\n");

  items=GTK_CLIST(whichside->list)->selection;
  while(items)
    {
      totsize+=((dentry *)gtk_clist_get_row_data(GTK_CLIST(whichside->list),(int)(items->data)))->size;
      items=items->next;
    }
  whichside->totsize=totsize;
  getpath(whichside,&path);
  drivepath=malloc((strlen(path)+30)*sizeof(char));
  sprintf(drivepath,"%li bytes selected in ",totsize);
  if(whichside->drive)
    sprintf(drivepath+strlen(drivepath),"%c:",whichside->drive);
  strcat(drivepath,path);
  gtk_statusbar_pop(GTK_STATUSBAR(whichside->status),whichside->context);
  gtk_statusbar_push(GTK_STATUSBAR(whichside->status),whichside->context,drivepath);
  free(path);
  free(drivepath);
  if(event&&(event->type==GDK_2BUTTON_PRESS))
    {
      gchar *text;
      if(debug)
	printf("  doubleclick\n");
      gtk_clist_unselect_row(GTK_CLIST(widget),row,column);
      gtk_clist_get_text(GTK_CLIST(widget),row,1,&text);
      if(((dentry *)gtk_clist_get_row_data(GTK_CLIST(whichside->list),row))->dir)
	{
	  if(strcmp(text,"..")&&(whichside->length<DIR_DEPTH-1))
	    {
	      whichside->path[whichside->length]=(gchar *)malloc((strlen(text)+1)*sizeof(gchar));
	      strcpy(whichside->path[whichside->length],text);
	      whichside->length+=1;
	    }
	  else if(whichside->length)
	    {
	      whichside->length-=1;
	      free(whichside->path[whichside->length]);
	    }
	  refresh_event(widget,whichside);
	}
    }
  if(debug)
    printf("End dclick_callback()\n");
  return(FALSE);
}

gint rclick_callback(GtkWidget *widget,GdkEventButton *event,side *whichside)
{
  if(debug)
    printf("In rclick_callback()\n");
  if(event&&(event->type==GDK_BUTTON_PRESS)&&(event->button==3))
    {
      gtk_menu_popup(GTK_MENU(whichside->contextmenu),
		     NULL,NULL,NULL,NULL,
		     event->button,event->time);
      return(TRUE);
    }
  else
    return(FALSE);
}

gint sort_callback(GtkWidget *widget,gint column,side *whichside)
{
  if(debug)
    printf("In sort_callback()\n");
  if(column>0)
    {
      gtk_clist_set_sort_column(GTK_CLIST(whichside->list),column);
      gtk_clist_sort(GTK_CLIST(whichside->list));
    }
  return(TRUE);
}

int datecompare(const char *text1,const char *text2)
{
  int comp;
  if (!text2)
    return (text1 != NULL);
  if (!text1)
    return -1;
  if((comp=memcmp(text1+6,text2+6,4))<0)
    return -1;
  else if(comp >0)
    return 1;
  if((comp=memcmp(text1,text2,2))<0)
    return -1;
  else if(comp >0)
    return 1;
  if((comp=memcmp(text1+3,text2+3,2))<0)
    return -1;
  else if(comp >0)
    return 1;
  return 0;
}

static gint sortnew(GtkCList *clist, gconstpointer ptr1, gconstpointer ptr2)
{
  /* copied from GTKs gtkclist.c and modified */
  if(clist->sort_column==2)
    {
      char *text1 = NULL;
      char *text2 = NULL;
      GtkCListRow *row1 = (GtkCListRow *) ptr1;
      GtkCListRow *row2 = (GtkCListRow *) ptr2;

      switch (row1->cell[clist->sort_column].type)
        {
        case GTK_CELL_TEXT:
	  text1 = GTK_CELL_TEXT (row1->cell[clist->sort_column])->text;
	  break;
        case GTK_CELL_PIXTEXT:
	  text1 = GTK_CELL_PIXTEXT (row1->cell[clist->sort_column])->text;
	  break;
        default:
	  break;
        }
      
      switch (row2->cell[clist->sort_column].type)
        {
        case GTK_CELL_TEXT:
	  text2 = GTK_CELL_TEXT (row2->cell[clist->sort_column])->text;
	  break;
        case GTK_CELL_PIXTEXT:
	  text2 = GTK_CELL_PIXTEXT (row2->cell[clist->sort_column])->text;
	  break;
        default:
	  break;
        }

      if (!text2)
	return (text1 != NULL);
      if (!text1)
	return -1;
      if(atol(text1)==atol(text2))
	return 0;
      else
	return atol(text1)>atol(text2);
    }
  else if((clist->sort_column==3)||(clist->sort_column==4))
    {
      char *textdate1 = NULL;
      char *textdate2 = NULL;
      char *texttime1 = NULL;
      char *texttime2 = NULL;
      GtkCListRow *row1 = (GtkCListRow *) ptr1;
      GtkCListRow *row2 = (GtkCListRow *) ptr2;

      int comp;

      switch (row1->cell[3].type)
        {
        case GTK_CELL_TEXT:
	  textdate1 = GTK_CELL_TEXT (row1->cell[3])->text;
	  break;
        case GTK_CELL_PIXTEXT:
	  textdate1 = GTK_CELL_PIXTEXT (row1->cell[3])->text;
	  break;
        default:
	  break;
        }
      
      switch (row2->cell[3].type)
        {
        case GTK_CELL_TEXT:
	  textdate2 = GTK_CELL_TEXT (row2->cell[3])->text;
	  break;
        case GTK_CELL_PIXTEXT:
	  textdate2 = GTK_CELL_PIXTEXT (row2->cell[3])->text;
	  break;
        default:
	  break;
        }
      switch (row1->cell[3].type)
        {
        case GTK_CELL_TEXT:
	  texttime1 = GTK_CELL_TEXT (row1->cell[4])->text;
	  break;
        case GTK_CELL_PIXTEXT:
	  texttime1 = GTK_CELL_PIXTEXT (row1->cell[4])->text;
	  break;
        default:
	  break;
        }
      
      switch (row2->cell[3].type)
        {
        case GTK_CELL_TEXT:
	  texttime2 = GTK_CELL_TEXT (row2->cell[4])->text;
	  break;
        case GTK_CELL_PIXTEXT:
	  texttime2 = GTK_CELL_PIXTEXT (row2->cell[4])->text;
	  break;
        default:
	  break;
        }

      if((comp=datecompare(textdate1,textdate2)))
	return comp;
      return(strcmp(texttime1,texttime2));
    }
  else
    return (*default_compare_func)(clist, ptr1, ptr2);
}
 
GtkWidget *create_side(side *whichside,char *copytext)
{
  GtkItemFactoryEntry menu_items[] = {
       { "/_File",                    NULL,         NULL,        0,       "<Branch>"       },
       { "/File/Delete",              NULL,         NULL,        0,       NULL             },
       { "/File/Rename",              NULL,         NULL,        0,       NULL             },
       { "/File/Print",               NULL,         NULL,        0,       NULL             },
       { "/File/Make directory",      NULL,         NULL,        0,       NULL             },
       { "/_View",                    NULL,         NULL,        0,       "<Branch>"       },
       { "/View/Refresh",             NULL,         NULL,        0,       NULL             },
       { "/_Select",                  NULL,         NULL,        0,       "<Branch>"       },
       { "/Select/All",               NULL,         NULL,        0,       NULL             },
       { "/Select/None",              NULL,         NULL,        0,       NULL             },
     };
  GtkItemFactory *item_factory;
  GtkWidget *popup_item;
  GtkWidget *box,*inbox,*menubar;
#ifndef OLDGTK
  GtkWidget *scroll;
#endif

  if(debug)
    printf("In create_side()\n");

  box=gtk_vbox_new(FALSE,5);

  /* menubar, combobox for drive */
  inbox=gtk_hbox_new(FALSE,0);
  
  item_factory=gtk_item_factory_new(GTK_TYPE_MENU_BAR,"<main>",NULL);
  gtk_item_factory_create_items (item_factory,sizeof(menu_items)/sizeof(menu_items[0]),menu_items,NULL);
  menubar=gtk_item_factory_get_widget(item_factory,"<main>");
  gtk_signal_connect(GTK_OBJECT(gtk_item_factory_get_widget(item_factory,"/File/Delete")),
		     "activate",GTK_SIGNAL_FUNC(filedelete_event),whichside);
  gtk_signal_connect(GTK_OBJECT(gtk_item_factory_get_widget(item_factory,"/File/Rename")),
		     "activate",GTK_SIGNAL_FUNC(filerename_event),whichside);
  gtk_signal_connect(GTK_OBJECT(gtk_item_factory_get_widget(item_factory,"/File/Print")),
		     "activate",GTK_SIGNAL_FUNC(fileprint_event),whichside);
  gtk_signal_connect(GTK_OBJECT(gtk_item_factory_get_widget(item_factory,"/File/Make directory")),
		     "activate",GTK_SIGNAL_FUNC(mkdir_event),whichside);
  gtk_signal_connect(GTK_OBJECT(gtk_item_factory_get_widget(item_factory,"/View/Refresh")),
		     "activate",GTK_SIGNAL_FUNC(refresh_event),whichside);
  gtk_signal_connect(GTK_OBJECT(gtk_item_factory_get_widget(item_factory,"/Select/All")),
		     "activate",GTK_SIGNAL_FUNC(selectall_event),whichside);
  gtk_signal_connect(GTK_OBJECT(gtk_item_factory_get_widget(item_factory,"/Select/None")),
		     "activate",GTK_SIGNAL_FUNC(selectnone_event),whichside);

  gtk_box_pack_start (GTK_BOX (inbox),menubar,TRUE,TRUE,5);

  whichside->combo=gtk_combo_new();
  whichside->comboglist=NULL;
#ifndef OLDGTK
  gtk_widget_set_usize(whichside->combo, 80,0);
#endif
  gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(whichside->combo)->entry),FALSE);
  set_combo_entries(whichside);
  if(!whichside->drive)
    gethdpath(whichside);
  gtk_box_pack_end(GTK_BOX(inbox),whichside->combo,FALSE,TRUE,5);

  gtk_widget_show(menubar);
  gtk_widget_show(whichside->combo);

  gtk_box_pack_start(GTK_BOX(box),inbox,FALSE,TRUE,0);
  gtk_widget_show(inbox);
  gtk_signal_connect(GTK_OBJECT(GTK_COMBO(whichside->combo)->entry),
		     "changed",GTK_SIGNAL_FUNC(combo_change),whichside);

  /* clist and scrollbars */
#ifndef OLDGTK
  scroll=gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC); 
  gtk_box_pack_start(GTK_BOX(box),scroll,TRUE,TRUE,0);
#endif

  whichside->list=gtk_clist_new (5);
  gtk_clist_column_titles_show(GTK_CLIST(whichside->list));
  gtk_clist_set_selection_mode(GTK_CLIST(whichside->list),GTK_SELECTION_EXTENDED);
  gtk_clist_column_titles_passive(GTK_CLIST(whichside->list));
#ifndef OLDGTK
  gtk_container_add(GTK_CONTAINER(scroll),whichside->list);
#else
  gtk_clist_set_policy (GTK_CLIST(whichside->list),GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start(GTK_BOX(box),whichside->list,TRUE,TRUE,0);
#endif
  gtk_clist_set_column_title(GTK_CLIST(whichside->list),0,(gchar *) "");
  gtk_clist_set_column_title(GTK_CLIST(whichside->list),1,(gchar *) "Filename");
  gtk_clist_set_column_title(GTK_CLIST(whichside->list),2,(gchar *) "Size");
  gtk_clist_set_column_title(GTK_CLIST(whichside->list),3,(gchar *) "Date");
  gtk_clist_set_column_title(GTK_CLIST(whichside->list),4,(gchar *) "Time");

#ifdef NOPIXMAPS
  gtk_clist_set_column_width(GTK_CLIST(whichside->list),0,0);
#else
  gtk_clist_set_column_width(GTK_CLIST(whichside->list),0,20);
#endif

  gtk_clist_set_column_width(GTK_CLIST(whichside->list),1,110);
  gtk_clist_set_column_width(GTK_CLIST(whichside->list),2,50);
  gtk_clist_set_column_width(GTK_CLIST(whichside->list),3,75);
  gtk_clist_set_column_width(GTK_CLIST(whichside->list),4,30);

  gtk_clist_set_column_justification(GTK_CLIST(whichside->list),2,GTK_JUSTIFY_RIGHT);
  gtk_clist_set_sort_column(GTK_CLIST(whichside->list),1);
  gtk_clist_set_auto_sort(GTK_CLIST(whichside->list),TRUE);

  gtk_signal_connect(GTK_OBJECT(whichside->list),"select_row",GTK_SIGNAL_FUNC(dclick_callback),whichside);
  gtk_signal_connect(GTK_OBJECT(whichside->list),"unselect_row",GTK_SIGNAL_FUNC(dclick_callback),whichside);

  gtk_signal_connect(GTK_OBJECT(whichside->list),"button-press-event",
		     GTK_SIGNAL_FUNC(rclick_callback),whichside);

  gtk_signal_connect(GTK_OBJECT(whichside->list),"click-column",
		     GTK_SIGNAL_FUNC(sort_callback),whichside);

  default_compare_func=GTK_CLIST(whichside->list)->compare;
  gtk_clist_set_compare_func(GTK_CLIST(whichside->list),sortnew);

  gtk_widget_show(whichside->list);
#ifndef OLDGTK
  gtk_widget_show(scroll);
#endif

  /* status bar */
  whichside->status=gtk_statusbar_new();
  gtk_box_pack_start(GTK_BOX(box),whichside->status,FALSE,TRUE,0);
  whichside->context=gtk_statusbar_get_context_id(GTK_STATUSBAR(whichside->status),"left window");
  gtk_statusbar_pop(GTK_STATUSBAR(whichside->status),whichside->context);
  gtk_statusbar_push(GTK_STATUSBAR(whichside->status),whichside->context,"");

  gtk_widget_show(whichside->status);
  gtk_widget_show(box);

  /* context menu */
  whichside->contextmenu=gtk_menu_new();
  popup_item=gtk_menu_item_new_with_label("Delete");
  gtk_menu_append(GTK_MENU(whichside->contextmenu),popup_item);
  gtk_signal_connect(GTK_OBJECT(popup_item),"activate",
		     GTK_SIGNAL_FUNC(filedelete_event),
		     whichside);
  gtk_widget_show (popup_item);
  popup_item=gtk_menu_item_new_with_label("Rename");
  gtk_menu_append(GTK_MENU(whichside->contextmenu),popup_item);
  gtk_signal_connect(GTK_OBJECT(popup_item),"activate",
		     GTK_SIGNAL_FUNC(filerename_event),
		     whichside);
  gtk_widget_show (popup_item);
  popup_item=gtk_menu_item_new_with_label("Print");
  gtk_menu_append(GTK_MENU(whichside->contextmenu),popup_item);
  gtk_signal_connect(GTK_OBJECT(popup_item),"activate",
		     GTK_SIGNAL_FUNC(fileprint_event),
		     whichside);
  gtk_widget_show (popup_item);
  popup_item=gtk_menu_item_new_with_label(copytext);
  gtk_menu_append(GTK_MENU(whichside->contextmenu),popup_item);
  gtk_signal_connect(GTK_OBJECT(popup_item),"activate",
		     GTK_SIGNAL_FUNC(copy_event),
		     whichside);
  gtk_widget_show (popup_item);
  popup_item=gtk_menu_item_new_with_label("Make directory");
  gtk_menu_append(GTK_MENU(whichside->contextmenu),popup_item);
  gtk_signal_connect(GTK_OBJECT(popup_item),"activate",
		     GTK_SIGNAL_FUNC(mkdir_event),
		     whichside);
  gtk_widget_show (popup_item);

  if(debug)
    printf("End create_side()\n");
  return(box);
}

int main(int argc,char *argv[])
{
  GtkWidget *window;
  GtkWidget *menubar;
  GtkWidget *box,*inbox,*middlebox;
  GtkWidget *button;
  GtkTooltips *tooltips;
#ifndef NOPIXMAPS
  GtkStyle *style;
#endif
  GdkColor dircolor,linkcolor;
  GtkItemFactoryEntry menu_items[] = {
       { "/_File",                         NULL,         NULL,           0,   "<Branch>"       },
       { "/File/Quit",                     "<control>Q", NULL,           0,   NULL             },
       { "/_Options",                      NULL,         NULL,           0,   "<Branch>"       },
       { "/Options/Show dot files",        "<control>D", NULL,           0,   "<CheckItem>"    },
       { "/Options/Text mode for copying", "<control>T", NULL,           0,   "<CheckItem>"    },
       { "/Options/Separator",             NULL,         NULL,           0,   "<Separator>"    },
       { "/Options/Configure mfm ...",     "<control>C", NULL,           0,   NULL             },
       { "/_Help",                         NULL,         NULL,           0,   "<LastBranch>"   },
       { "/_Help/Command line options",    NULL,         command_event,  0,   NULL             },
       { "/_Help/How to use",              NULL,         use_event,      0,   NULL             },
       { "/_Help/About",                   NULL,         about_event,    0,   NULL             },
     };
  GtkItemFactory *item_factory;
  GtkAccelGroup *accel_group;

  side left,right;

  char *cfname;
  int i,nosysconf=0,nouserconf=0,aflag=1,bflag=1;

  gtk_init (&argc, &argv);

  putenv("MTOOLS_DATE_STRING=mm-dd-yyyy");
  putenv("MTOOLS_TWENTY_FOUR_HOUR_CLOCK=1");

  for(i=1;i<argc;i++)
    {
      if(!strcmp(argv[i],"--nosysconf"))
	nosysconf=1;
      else if(!strcmp(argv[i],"--nouserconf"))
	nouserconf=1;
      else if(!strcmp(argv[i],"--save"))
	nosave=0;
      else if(!strcmp(argv[i],"--nosave"))
	nosave=1;
      else if(!strcmp(argv[i],"-v"))
	debug=1;
      else if(!strcmp(argv[i],"--debug"))
	debug=1;
      else if(!strcmp(argv[i],"--help"))
	{
	  g_print("\n" ABOUTMFM "\n" MFMOPTIONS "\n");
	  exit(0);
	}
      else
	{
	  g_print("\nUnknown option %s\n",argv[i]);
	  g_print("\n" MFMOPTIONS "\n");
	  exit(0);
	}
    }
  left.other=&right;
  right.other=&left;
  if(debug)
    {
      printf("In main():\n");
      printf("  left : %p\n",&left);
      printf("  right: %p\n",&right);
      printf("  left.other : %p\n",left.other);
      printf("  right.other: %p\n",right.other);
    }
  left.length=0;
  right.length=0;
  left.drive='\0';
  right.drive='\0';
  strcpy(mpath,MTOOLSDIR);
  left.comboglist=NULL;
  right.comboglist=NULL;
  if(!nosysconf)
    aflag=read_configfile(&left,CONFFILE);
  homedir=g_get_home_dir();
  cfname=malloc((strlen(homedir)+6)*sizeof(char));
  strcpy(cfname,homedir);
  strcat(cfname,"/.mfm");
  if(!nouserconf)  
    bflag=read_configfile(&left,cfname);
  if(!(*mpath))
    strcpy(mpath,MTOOLSDIR);
  if(aflag&&bflag)
    strcpy(drives,"a");
  if(debug)
    {
      printf("  mpath: %s+\n",mpath);
      printf("  cfname: %s+\n",cfname);
      printf("  drives: %s+\n",drives);
    }
  free(cfname);

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), "mfm " VERSION);
  gtk_widget_set_usize(GTK_WIDGET(window), 750, 400);
  gtk_signal_connect(GTK_OBJECT(window),"delete_event",GTK_SIGNAL_FUNC(delete_event),&left);

  box=gtk_vbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(window),box);

  tooltips=gtk_tooltips_new ();
  
  /* menu */
  accel_group=gtk_accel_group_new();
  item_factory=gtk_item_factory_new(GTK_TYPE_MENU_BAR,"<main>",accel_group);
  gtk_item_factory_create_items (item_factory,sizeof(menu_items)/sizeof(menu_items[0]),menu_items,NULL);
  gtk_accel_group_attach(accel_group,GTK_OBJECT(window));
  menubar=gtk_item_factory_get_widget(item_factory,"<main>");
  gtk_signal_connect(GTK_OBJECT(gtk_item_factory_get_widget(item_factory,"/File/Quit")),
		     "activate",GTK_SIGNAL_FUNC(quit_event),&left);
  gtk_signal_connect(GTK_OBJECT(gtk_item_factory_get_widget(item_factory,"/Options/Configure mfm ...")),
		     "activate",GTK_SIGNAL_FUNC(config_event),&left);
  showdots=gtk_item_factory_get_widget(item_factory,"/Options/Show dot files");
  textmode=gtk_item_factory_get_widget(item_factory,"/Options/Text mode for copying");
  gtk_signal_connect(GTK_OBJECT(showdots),"toggled",GTK_SIGNAL_FUNC(refresh_event),&left);
  gtk_signal_connect(GTK_OBJECT(showdots),"toggled",GTK_SIGNAL_FUNC(refresh_event),&right);
  gtk_box_pack_start (GTK_BOX(box),menubar,FALSE,TRUE,0);
  gtk_widget_show(menubar);

  inbox=gtk_hbox_new(FALSE,0);
#ifndef OLDGTK
  gtk_container_set_border_width(GTK_CONTAINER(inbox),5);
#endif
  gtk_box_pack_start(GTK_BOX(box),inbox,TRUE,TRUE,0);  
  gtk_widget_show(box);

  /* left */
  gtk_box_pack_start(GTK_BOX(inbox),create_side(&left,"Copy >>>"),TRUE,TRUE,0);

  /* middle */
  middlebox=gtk_vbox_new(FALSE,10);
  gtk_box_pack_start(GTK_BOX(inbox),middlebox,FALSE,TRUE,10);

  button=gtk_button_new_with_label ("->\nCopy");
  gtk_tooltips_set_tip(tooltips,button,"Copy file(s) from left to right window",NULL);
  gtk_box_pack_start(GTK_BOX(middlebox),button,TRUE,FALSE,10);
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",GTK_SIGNAL_FUNC(copy_event),&left);

  button=gtk_button_new_with_label ("Copy\n<-");
  gtk_tooltips_set_tip(tooltips,button,"Copy file(s) from right to left window",NULL);
  gtk_box_pack_start(GTK_BOX(middlebox),button,TRUE,FALSE,10);
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button),"clicked",GTK_SIGNAL_FUNC(copy_event),&right);

  gtk_widget_show(middlebox);

  /* right */
  gtk_box_pack_end(GTK_BOX(inbox),create_side(&right,"<<< Copy"),TRUE,TRUE,0);

  gtk_widget_show(inbox);

  /* allocate colors */
  dircolor.red=0;
  dircolor.green=0;
  dircolor.blue=65535;
  dircolor.pixel =(gulong)255;
  gdk_color_alloc(gtk_widget_get_colormap(window),&dircolor);
  left.dircolor=&dircolor;
  right.dircolor=&dircolor;
  linkcolor.red=98*255;
  linkcolor.green=170*255;
  linkcolor.blue=78*255;
  linkcolor.pixel =(gulong)255;
  gdk_color_alloc(gtk_widget_get_colormap(window),&linkcolor);
  left.linkcolor=&linkcolor;
  right.linkcolor=&linkcolor;

  /*  show all */
  gtk_widget_show(window);

#ifndef NOPIXMAPS
  style=gtk_widget_get_style(window);
  dirpix=gdk_pixmap_create_from_xpm_d(window->window,&dirmask,&style->bg[GTK_STATE_NORMAL],(gchar **)xpm_dir);
  filepix=gdk_pixmap_create_from_xpm_d(window->window,&filemask,&style->bg[GTK_STATE_NORMAL],(gchar **)xpm_file);
  linkdirpix=gdk_pixmap_create_from_xpm_d(window->window,&linkdirmask,&style->bg[GTK_STATE_NORMAL],(gchar **)xpm_linkdir);
  linkfilepix=gdk_pixmap_create_from_xpm_d(window->window,&linkfilemask,&style->bg[GTK_STATE_NORMAL],(gchar **)xpm_linkfile);
  mfmpix=gdk_pixmap_create_from_xpm_d(window->window,&mfmmask,&style->bg[GTK_STATE_NORMAL],(gchar **)xpm_mfm);
  gdk_window_set_icon(window->window,NULL,mfmpix,mfmmask);
#endif

  refresh_event(left.list,&left);
  refresh_event(right.list,&right);

  while (gtk_events_pending())
    gtk_main_iteration();

  gtk_main ();
  if(debug)
    printf("End main()\n");
  return(0);
}

